#
# -*- Mode:Perl; -*-
#

# invalidDirectivesSeen records where such directives were seen
%invalidDirectivesSeen = ();
$invalidDirectivesReason = "The C language defines the set of preprocessor directives that may be used.  Some compilers support other directives, but programs that take advantage of those directives are not portable.";

# This is used to keep track of all of the names that are not "good"
%invalidNamesSeen = ();

# List the known C preprocessor directives
# Note particularly the "warning" is not in this list; warning is not
# standard.  Ditto for abort.  This list contains *ALL* of the valid
# and portable directive names.  DO NOT ADD TO IT.
%allowedDirectives = ( 'if' => 1,
		       'ifdef' => 1,
		       'ifndef' => 1,
		       'else' => 1,
		       'elif' => 1,
		       'endif' => 1,
		      'pragma' => 1,
		      'define' => 1,
		      'undef' => 1, 
		      'error' => 1,
		      'include' => 1,
		      );

# gAllowedDefines are the defines that are always allowed
# (this is a subset of the valid names)
%gAllowedDefines = ( 'SA_RESETHAND' => 'sys', 
		     'WORDS_BIGENDIAN' => 'autoconf',
		     'EAGAIN' => 'sys',
		     'ENOENT' => 'sys',
		     'EIDRM'  => 'sys',
		     'EPERM'  => 'sys',
		     'SEEK_SET' => 'sys',
		     'SEEK_CUR' => 'sys',
		     'SEEK_END' => 'sys',
		     'NULL' => 'sys',
		     'SI_SRPC_DOMAIN' => 'sys',
		     'NDEBUG' => 'sys',         # defined for assert()
		     );

# Code for allowing define changes to be properly nested 
# within directory trees.
# An entry on this stack contains
#    dirdepth:variablename:scope
# where dirdepth is the directory depth where the variable was added
# (this tells when to pop the item), the variable name is a string
# that names the hash for the allowed defines, and scope is either
# replace or add, and indicates whether this adds to or replaces the 
# variables below this point on the stack.  (e.g., we stop when
# we get to a replace).
@gAllowedDefinesForDirStack = ();
@gAllowedDefinesForDirTreeStack = ();

# AllowedDefines are the defines allowed for this file only; it is initialized
# from gAllowedDefines and gAllowedDefinesForDirStack
%AllowedDefines = ();
@GoodNames = ( 'HAVE_[A-Za-z0-9_]*',  'USE_[A-Za-z0-9_]*',
	       'NEEDS_[A-Za-z0-9_]*', 'WITH_[A-Z0-9_]*',

	       'STDC_HEADERS', '__cplusplus', 
	       'IOV_MAX', 'S_ISLNK', 
	       'SA_RESTART', 'SA_INTERRUPT', # Sigaction values
	       'SIGINT', 'SIGALRM', 'SIGQUIT', # Add other signals as needed
	       'SIGSYS', 'SIGFPE', 'SIGIO', 'SIGSEGV', 'SIGSTOP',
	       'SIGCONT',
               'MAXHOSTNAMELEN',             # Unix sys define for host names
	       'MAXPATHLEN',                 # Unix define for path elements
	       'ADDR_ANY', 'INADDR_ANY',     # network defines
	       'MPICH_SINGLE_THREADED', 'CHAR_PTR_IS_ADDRESS', 'FOO',
	       'MPIR_[A-Za-z0-9_]*',
	       'MPID_[A-Za-z0-9_]*',
	       'MPIDU_[A-Za-z0-9_]*',   # More internal MPID names
	       'MPIU_[A-Za-z0-9_]*',
	       'MPIDI_[A-Za-z0-9_]*',   # Internal MPID names
	       'MPICH_[A-Za-z0-9_]*', 
	       'MPIO_[A-Za-z0-9_]*', 
	       'DBG_[A-Z0-9_]*',
	       '_H_INCLUDED',           # Ok for header files
	       'DEBUG_[A-Z0-9_]*',
	       'F77_[A-Z_]*',
	       'MPI_BUILD_PROFILING',
	       'MPI_Info_f2c', 'MPI_Info_c2f', 'MPI_File_f2c', 'MPI_File_c2f', 
	       'MPI_DUP_FN',            # protected in f77 interface
	       'AFX_.*_INCLUDED_',      # Autogenerated defines for some Windows code
	       'APSTUDIO_INVOKED',      # Autogenerated ?
	       'MPI_COMM_DUP_FN',       # The following 9 definitions
	       'MPI_TYPE_DUP_FN',       # are used to handle the attribute
	       'MPI_WIN_DUP_FN',        # dup functions
	       'MPI_COMM_NULL_COPY_FN', 
	       'MPI_COMM_NULL_DELETE_FN',
	       'MPI_TYPE_NULL_COPY_FN', 
	       'MPI_TYPE_NULL_DELETE_FN',
	       'MPI_WIN_NULL_COPY_FN', 
	       'MPI_WIN_NULL_DELETE_FN',
	       'MPI_CONVERSION_FN_NULL',
	       'MPI_SEEK_SET',          # These two are used in the C++ 
	       'MPI_MODE_RDONLY',       # interface to check for problems
	       );

@BadNames = ( 'rs6000', 'solaris', 'hpux', 'linux', 'HPUX', 'IRIX', 'LINUX', );

# ---------------------------------------------------------------------------
sub ClearDefines {
    my $filename = $_[0];
    %AllowedDefines = %gAllowedDefines;

    for (my $i=$#gAllowedDefinesForDirStack; $i>=0; $i--) {
	my $elm = $gAllowedDefinesForDirStack[$i];
	my ($depth,$varname,$action) = split(/:/,$elm);
	
	foreach my $key (keys(%$varname)) {
	    $AllowedDefines{$key} = $$varname{$key};
	}
	if ($action eq "replace") { last; }
    }

    for (my $i=$#gAllowedDefinesForDirTreeStack; $i>=0; $i--) {
	my $elm = $gAllowedDefinesForDirTreeStack[$i];
	my ($depth,$varname,$action) = split(/:/,$elm);
	
	foreach my $key (keys(%$varname)) {
	    $AllowedDefines{$key} = $$varname{$key};
	}
	if ($action eq "replace") { last; }
    }

    %defsUsed = ();
}

sub ClearDirDefines {
    if ($#gAllowedDefinesForDirStack >= 0) {
	my $elm = $gAllowedDefinesForDirStack[$#gAllowedDefinesForDirStack];
	my ($depth,$varname,$action) = split(/:/,$elm);
	if ($depth == $gDirDepth) {
	    $#gAllowedDefinesForDirStack --;
	}
    }
}
sub ClearDirTreeDefines {
    # Clear when we exit the same directory tree we entered
    # Eventually, we need a stack for these
    if ($#gAllowedDefinesForDirTreeStack >= 0) {
	my $elm = $gAllowedDefinesForDirTreeStack[$#gAllowedDefinesForDirTreeStack];
	my ($depth,$varname,$action) = split(/:/,$elm);
	if ($depth == $gDirDepth) {
	    $#gAllowedDefinesForDirTreeStack --;
	}
    }
}

sub CheckLineDefines {
    my ($FD, $curline, $filename) =@_;


    # Only apply to C or CXX files, or to files that have set
    # $gIsCPPFile (e.g., foo.F)
    if ( $gLanguage ne "CXX" && $gLanguage ne "C" && $gIsCPPFile == 0) {
	return $curline; }

    if ($curline =~ /^#/) {
	print "Checking $curline for bad directives\n" if $gDebug;
	$curline = &CheckForBadDirectives( $FD, $curline, $filename );
	if ($checkDefines) {
	    print "Checking $curline for bad defines\n" if $gDebug;
	    $curline = &CheckForGuardedDefines( $FD, $curline, $filename );
	    $curline = &CheckForBadDefines( $FD, $curline, $filename );
	    $curline = &CheckForBadDefinitions( $FD, $curline, $filename );
	}
    }
    return $curline;
}

# Summarize the reason for any invalid directive
sub FinalReportOnDefines {
    my $headerPrinted = 0;
    foreach my $key (keys(%invalidDirectivesSeen)) {
	if (!$headerPrinted) {
	    $headerPrinted = 1;
	    print "$topHeader\n$invalidDirectivesReason\n";
	    print "The following invalid directives were encountered:\n";
	    $topHeader = "";
	}
	print "$key\n";
    }

    $headerPrinted = 0;
    foreach my $key (sort(keys(%invalidNamesSeen))) {
	if (!$headerPrinted) {
	    $headerPrinted = 1;
	    print "The following invalid ifdef names were encountered:\n";
	    $topHeader = "";
	}
	print "$key\n";
    }
}
# ---------------------------------------------------------------------------
# Routines to allow customization of the names that are allowed
# ( name of hash containing names, tree or dir, add or replace ), 
# eg., 
#   %mynames = ( 'MYPACK_.*' => mypack );
#   &PushDefinesNames( "mynames", "tree", "replace" );
sub PushDefinesNames {
    my ($varname, $dirscope, $action ) = @_;
    my $elm;

    if ($action ne "add" && $action ne "replace") {
	print STDERR "Action in PushDefinesNames is $action, must be either add or replace";
	return 0;
    }
    $elm = $gDirDepth . ":" . $varname . ":" . $action;
    
    if ($dirscope eq "tree") {
	$gAllowedDefinesForDirTreeStack[$#gAllowedDefinesforDirTreeStack+1] = 
	    $elm
    }
    elsif ($dirscope eq "dir") {
	$gAllowedDefinesForDirStack[$#gAllowedDefinesforDirStack+1] = 
	    $elm
    }
    else {
	print STDERR "Invalid scope in PushDefinesNames; was $dirscope but must be either tree or dir\n";
	return 0;
    }
    return 1;
}

# This routine is called by the style checker
sub Styledefine {
    my ($action,$name,$count,$sig) = @_;

    # There should be special support for style and for structured comments
    # to pass information to other modules, perhaps with the style name
    # as the key.
    #print "Allowing define of :$name:\n"; 
    $AllowedDefines{$name} = $count;
}

# ---------------------------------------------------------------------------
# Helper routines
sub CheckDefineName {
    my $name = $_[0];
    
    # Remove blanks from around name
    $name =~ s/\s//g;

    foreach $good (@GoodNames) {
	if ($name =~ /$good/) { return 0; }
    }

    foreach $bad (@BadNames) {
	if ($name =~ /$bad/) { return 1; }
    }
    if ($name eq "__DATE__") {
	# We allow this because some older C compilers don't set it.
	# We should really check that it is not set withing an 
	#  #ifndef __DATE__... 
	return 0;
    }

    # Check for defining a __ name (not allowed by ``users'', only
    # by the environment.  In the case of GCC, instead of setting
    # things like __USE_MISC, you should set the appropriate _xxx_SOURCE
    # variable.)
    if ($name =~ /^__.*/) { return 1; }

    return 1;
}

# In the process for checking for bad definitions, we also *remember* all
# of the definitions that we see.  This helps with the tests for 
# guarded routines
sub CheckForBadDefines {
    my $FD = $_[0];
    my $curline = $_[1];
    my $filename = $_[2];
    if ($curline =~ /^\s*#if/) {
	# Read the continuation lines
	while ($curline =~ /\/$/) {
	    $curline .= <$FD>;
	    $line_num ++;
	}
	# Search for either ifdef or defined(...)
	if ($curline =~ /^\s*#if(n?)def\s+(\S+)\s*/) {
	    $isndef = $1;
	    $defname = $2;
	    $defname =~ s/\r*\n*//g;
	    $defsUsed{$defname} = "$filename:$curline";
	    print "checking #ifdef name $defname\n" if $gDebug;
	    if (! defined($AllowedDefines{$defname}) && 
		&CheckDefineName( $defname ) ) {
		&PrintCurrentDirectory;
		print "Warning: if${isndef}def name $defname used in $filename is nonstandard\n";
		$gAvoidConstructionsSeen{"define"} .= "$filename;";
		$invalidNamesSeen{$defname} = 1;
	    }
	}
	elsif ($curline =~ /^\s*#if\s/ || $curline =~ /^\s*#elif\s/) {
	       while ($curline =~ /defined\(\s*([^\)\s]*)\s*\)/) {
		   $defname = $1;
		   $defsUsed{$defname} = "$filename:$curline";
		   print "$defname\n" if $gDebug;
		   if (! defined($AllowedDefines{$defname}) && 
		       &CheckDefineName( $defname ) ) {
		       &PrintCurrentDirectory;
		       print "Warning: ifdef name $defname used in $filename is nonstandard\n";
		       $gAvoidConstructionsSeen{"define"} .= "$filename;";
		       $invalidNamesSeen{$defname} = 1;
		       }
		   $curline =~ s/defined\(\s*$defname\s*\)//g;
	       }
	}
    }
    return $curline;
}
# This checks for definitions that the user should not make (e.g., 
# #define __foo)
sub CheckForBadDefinitions {
    my $FD = $_[0];
    my $curline = $_[1];
    my $filename = $_[2];
    if ($curline =~ /^#\s*define\s+(__\w*)/) {
	$defname = $1;
	if ($defname eq "__DATE__") {
	    # We allow this because some older C compilers don't set it.
	    # We should really check that it is not set withing an 
	    #  #ifndef __DATE__... 
	    return $curline
	}
	print "Warning: definition of name $defname used in $filename is not allowed\n";
	$gAvoidConstructionsSeen{"define__"} .= "$filename;";
    }
    return $curline;
}

sub CheckForBadDirectives{
    my $FD = $_[0];
    my $curline = $_[1];
    my $filename = $_[2];
    if ($curline =~ /^#\s*(\w+)/) {
	$directiveName = $1;
	if (!defined($allowedDirectives{$directiveName})) {
	    print "Warning: directive $directiveName is unknown in $filename\n";
	    $invalidDirectivesSeen{$directiveName} .= "$filename;";
	}
    }
    return $curline;
}

#
# This checks for
# # ifdef NEEDS_foo_DEFINED
# # define foo
# # endif
# Basically, it allows you to skip over a define that would otherwise
# cause an error report in CheckForBadDefines
sub CheckForGuardedDefines {
    my $FD = $_[0];
    my $curline = $_[1];
    my $filename = $_[2];
    my $allowed_name = "";
    if ($curline =~ /^#\s*ifdef\s+NEEDS_(.*)_DEFINED/ ||
	$curline =~ /^#\s*if\s+defined\(NEEDS_(.*)_DEFINED\)/) {
        $allowed_name = $1;
	# Read *only* the next line
	$curline = <$FD>;
	# Allow some extra underscores before and after 
	if ($curline =~ /^#\s*define\s+_*${allowed_name}_*/) {
	    ;
	}
	else {
	    &PrintCurrentDirectory;
	    print "Warning: unexpected line after NEEDS_${allowed_name}_DEFINED in $filename\n";
        }
    }
    return $curline;
}

# Add to known actions
$KnownActions[$#KnownActions+1] = "Defines";

$gAvoidConstructionsReasons{"define"} = "\
Names defined for the C preprocessor (i.e., with #define) should be clearly
distinct from any name that may be defined by another, outside module, such as
a system or third-party header file.  Unfortunately, system headers do not
follow this rule (one system defined CACHE_SIZE), so we must.";

$gAvoidConstructionsReasons{"define__"} = "\
The C standard restricts the use of names with two leading underscores to
\"the system\"; these names should not be used in applications.  The intent
is to allow the compiler to have a private name space of preprocessor names";

sub DescribeDefines {
    print STDOUT 
	"\tcppdefines - Check C Preprocessor usage, including names and directives\n";
}

# This is currently too MPICH specific.  The various MPI-specific names
# should be placed within an mpich module.

1;
