#
# -*- Mode:Perl; -*-
#
# This module supports the MPICH2 test for nested routines
#
$notInNested = 1;
$nestedSeen  = 0;

sub ClearNested {
    $notInNested = 1;
    $nestedSeen  = 0;
}

sub CheckLineNested {
    my ($FD, $curline, $filename) =@_;

    if ($curline =~ /\/\*\s*begin:nested\s*\*\//) {
	$notInNested = 0;
    }
    elsif ($curline =~ /\/\*\s*end:nested\s*\*\//) {
	$notInNested = 1;
    }

    # We also need to check for the use of MPI routines from within
    # the code (only use NMPI), at least for all of the MPI
    # implementation code.  Withing src/mpi/*, we could look for
    # MPI_([A-Z][a-z0-9_]*)[^;]*; and then complain if $1 is not one
    # of Status, Comm, Group, Win, File, Aint, Fint, Offset, Request,
    # File, Info, Datatype .
    if ($notInNested) {
	&CheckForNested( FD, $_, $filename );
    }

    return $curline;
}

#
# Check for nested routine calls (NMPI) without Nest_incr/decr
# Special feature:
# /* begin:nested */ and /* end:nested */ can be used around
# routines that are only called with by another routine that
# has incremented the nested count

sub CheckForNested {
    my $FD = $_[0];
    my $curline = $_[1];
    my $filename = $_[2];
    
    # Skip definitions
    if ($curline =~ /#define\s+NMPI/) {
	return;
    }
    if ($curline =~ /NMPI_H_INCLUDED/) { 
	return;
    }
    # Check for usage
    if ($curline =~ /NMPI_([A-Za-z0-9_]*)/) {
	my $name = $1;
	if (!$nestedSeen && $name ne "Abort") {
	    # Don't warn about NMPI_Abort
	    &PrintCurrentDirectory;
	    print "Error: NMPI_$name called without MPIR_Nest_incr in $filename\[$line_num\]\n";
	}
    }
    elsif ($curline =~ /MPIR_Nest_incr/) {
	$nestedSeen ++;
    }
    elsif ($curline =~ /MPIR_Nest_decr/) {
	# We do not decrement nested seen since we do not analyze program 
	# structure
	# $nestedSeen --;
	;
    }
}

# Add to known actions
$KnownActions[$#KnownActions+1] = "Nested";

1;
