#
# -*- Mode:Perl; -*-
#
$doCheckForPMPI = 0;
%AllowPMPIRoutines = ();

sub ClearPMPI {
    my $filename = $_[0];

    $doCheckForPMPI = 0;
    %AllowPMPIRoutines = ();
    # Allow PMPI calls within the f77 binding and the nmpi.h file
    if ( (! ($filename =~ /binding\/f77/)) &&
	 (! ($filename =~ /include\/nmpi\.h/)) &&
	 (! ($filename =~ /include\/mpi\.h/)) &&
	 (! ($filename =~ /include\/mpif\.h/)) &&
	 (! ($filename =~ /include\/mpio\.h/)) &&
	 (! ($filename =~ /src\/util\/multichannel\/mpi\.c/)) &&
	 (! ($filename =~ /src\/mpe2/)) &&
	 (! ($filename =~ /romio\/adio\/include\/adio.h/)) &&
	 (! ($filename =~ /romio\/include\/mpio.h/)) &&
	 (! ($filename =~ /romio\/mpi-io\/fortran/)) &&
	 (! ($filename =~ /romio\/mpi2-other\/array\/fortran/)) &&
	 (! ($filename =~ /romio\/test/)) &&
	 (! ($filename =~ /romio\/mpi2-other\/info\/fortran/))
	 ){
	$doCheckForPMPI = 1;
    }
}

# Check for PMPI routine calls (normally, all calls should use NMPI)
sub CheckLinePMPI {
    my ($FD, $curline, $filename) =@_;
    
    if (! $doCheckForPMPI) { return $curline; }

    # Skip definitions
    if ($curline =~ /#define\s+MPI/) { # Skip #define MPI_xxx PMPI_xxx
	return;
    }
    if ($curline =~ /pragma/) { # Skip pragmas containing PMPI
	return;
    }
    if ($curline =~ /PMPI_LOCAL/) { # Skip the PMPI_LOCAL defn
	return;
    }
    if ($curline =~ /PMPI_([A-Za-z0-9_]*)/) {
	$name     = $1;
	$fullname = "PMPI_$name";
	if (defined($AllowPMPIRoutines{$fullname})) {
	    $AllowPMPIRoutines{$fullname}--;
	    if ($AllowPMPIRoutines{$fullname} >= 0) {
		return $curline;
	    }
	}
	&PrintCurrentDirectory;
	print "Error: PMPI_$name called in $filename\[$line_num\]\n";
    }
    return $curline;
}

sub StylePMPIuse {
    my ($action,$name,$count,$sig) = @_;
    #print "Setting allow for $name to $count\n";

    $AllowPMPIRoutines{$name} = $count;
}


# Add to known actions
$KnownActions[$#KnownActions+1] = "PMPI";

# Add to known File Extensions
# (.i files are C files that are included depending on configure options
$KnownFileExtensions{'i'} = 'C';

# Also make some other packages understand MPICH2 specific features
# cppdefines contains special features for Romio
#$allowRomioNames = 1;
#
# # We need a way to add these names *after* the module is loaded
sub PostLoad {
    $GoodNames[$#GoodNames+1] = 'ROMIO_[A-Za-z0-9_]*';
    
    # Make sure that the comments action is BEFORE the PMPI action.
    # This is a hack for now; we need a better way to separate out
    # pre and post comments actions.
    my $commentsLoc = -1;
    my $PMPILoc      = -1;
    for (my $i=0; $i<=$#KnownActions; $i++) {
	if ($KnownActions[$i] eq "PMPI") { $PMPILoc = $i; }
	if ($KnownActions[$i] eq "Comments") { $commentsLoc = $i; }
    }
    if ($commentsLoc > $PMPILoc && $PMPILoc >= 0) {
	# Swap them
	$KnownActions[$PMPILoc]     = "Comments";
	$KnownActions[$commentsLoc] = "PMPI";
    }
}

1;

