#
# This module will contain code to keep track of the variables that 
# have been defined for configure and that have been initialized to 
# a make variable

# Register (make-var-name, config-var-name)
# e.g., use &ConfigureRegister( "install", "\@INSTALL\@" )
# if the configure variable @INSTALL@ is used to initialize the make
# variable install.
#
# Functions Exported:
#  ConfigureRegiserDirUsed( autoconf name of directory )
#
# Variables Imported:
#  gAutoconfArgs
#
$KnownActions[$#KnownActions+1] = "Configure";

%ConfigureMakeVars = ();
%MakeConfigureVars = ();

# Temporary until configure reading code is fixed.
$gConfigureDir = "Unknown";

# This variable contains arguments that are used for all autoconf (and
# autoheader invocation).  Typically contains the arguments to 
# specify the directory containing macro defintions, using "ROOTDIR" 
# to allow the path to be adjusted automatically for any configure 
# within the project's directory tree
$gAutoconfArgs = "";

# Directories needed.  These are the prefix etc.  Many are needed only
# at the top level, particularly for installation.  srcdir is needed only for
# VPATH builds
# We must build the srcdir and top_srcdir specially because autoconf
# insists on replacing this text.  The rest we had to do to allow the
# maint/configure to work under Cygwin
%ConfigureDirsUsed = ();
%ConfigDirnames = (
		   srcdir           => "\@srcdir\@", 
		   master_topsrcdir => "\@master_top_srcdir\@",
		   topsrcdir        => "\@top_srcdir\@",
		   prefix           => "\@prefix\@",
		   exec_prefix      => "\@exec_prefix\@",
		   bindir           => "\@bindir\@",
		   sbindir          => "\@sbindir\@",
		   libdir           => "\@libdir\@",
		   includedir       => "\@includedir\@",
		   mandir           => "\@mandir\@",
		   htmldir          => "\@htmldir\@",
		   docdir           => "\@docdir\@",
		   sysconfdir        => "\@sysconfdir\@",
		   );

# These give the directories that autoconf will require for some of the
# directory (e.g., many are defined in terms of prefix or exec prefix)
%ConfigureRequiredDirs = ( 'libdir'      => 'exec_prefix prefix',
			   'bindir'      => 'exec_prefix prefix',
			   'mandir'      => 'prefix',
			   'htmldir'     => 'prefix',
			   'docdir'      => 'prefix',
			   'includedir'  => 'prefix',
			   'exec_prefix' => 'prefix',
			   'sysconfdir'  => 'prefix',
			   );
# ----------------------------------------------------------------------------
# Exported Functions
sub ConfigureRegister {
    my ($varname, $configVar) = @_;

    $ConfigureMakeVars{$configVar} = $varname;
    $MakeConfigureVars{$varname}  = $configVar;
}
# Register a directory as in use by this makefile
sub ConfigureRegisterDirUsed {
    my $dirname = $_[0];
    
    $ConfigureDirsUsed{$dirname} = 1;
    print "Remembering to define directory $dirname\n" if $gDebug;
}
# ----------------------------------------------------------------------------
# Standard output routines
sub OutputHeaderConfigure {
    my $FD = $_[0];

    print $FD "# \@configure_input\@\n";
    print $FD "SHELL           = \@SHELL\@\n";

    # Print the directories and needed autoconf names
    # Question: how do we communicate that information from the module
    # that makes use of the rule.  Should it register any autoconf variables?
    #   ... yes!
    # More precisely, those modules should register any "outside" make
    # variables with the rule to construct them.  Then the autoconf
    # module can figure out how to separate them

    %dirAdded = ();   # Keep track of the directories that we've added
    foreach my $dir (sort(keys(%ConfigureDirsUsed))) {
	print "Adding definition for directory $dir\n" if $gDebug;
	# Does this dir have a prereq?
	if (defined($ConfigureRequiredDirs{$dir})) {
	    foreach my $predir (split(/\s+/,$ConfigureRequiredDirs{$dir})) {
		if (!defined($dirAdded{$predir})) {
		    print $FD "$predir = $ConfigDirnames{$predir}\n";
		    $dirAdded{$predir} = 1;
		}
	    }
	}
	# Now, handle this directory
	if (!defined($dirAdded{$dir})) {
	    print $FD "$dir = $ConfigDirnames{$dir}\n";
	    $dirAdded{$dir} = 1;
	}
    }

    foreach my $makevar (sort(keys(%MakeConfigureVars))) {
	my $value = $MakeConfigureVars{$makevar};
	print $FD "$makevar\t= $value\n";
    }
}

# output the configure-releated dependencies
# Qeustion: should this be a higher-level operation within sm (e.g.,
# an explicit OutputDependencies<module>?)
sub OutputConfigure {
    my $FD = $_[0];
    
    # FIXME: need to set gConfigdir and get gCurdir
    if ($gConfigureDir eq "") { $gConfigureDir = "."; }

    # Compute the relative path to gConfigureDir
    my ($relpath,$destpath) = &SMGetRelativePathToDir( $gConfigureDir );
    $relpath =~ s/\/$//;   # Remove any trailing slash
    if ($relpath eq "") { $relpath = "."; }
    if ($destpath) { $destpath =~ s/\/$//; }  # Remove any trailing slash
    if (!$destpath || $destpath eq "") { $destpath = "."; }
    print $FD "Makefile: Makefile.in $relpath/configure\n";
    print $FD "\t-if [ -x $relpath/config.status ] ; then \\\n";
    print $FD "\tCONFIG_FILES=$destpath/Makefile CONFIG_HEADERS= \$(SHELL) $relpath/config.status ; fi\n";
    print $FD "\n";

    if (-s "configure.in") {
	# FIXME: complete dependencies (see ConfigureAnalyze)
	print $FD "configure: configure.in\n";
	# Starting with autoconf 2.57, there is a "cahce" that like
	# all autoconf caches, isn't reliable and must be removed.
	print $FD "\t\@if [ -d autom4te.cache ] ; then rm -rf autom4te.cache ; fi\n";
	my $aargs = $gAutoconfArgs;
	# Replace ROOTDIR with the (relative) path to the root if needed
	if ($aargs =~ /ROOTDIR/) {
	    my $rootpath = &RootDirGetPathToRootdir();
	    $aargs =~ s/ROOTDIR/$rootpath/g;
	}
	# FIXME: Only generate autoheader if necessary
	print $FD "\tautoheader $aargs\n\tautoconf $aargs\n";

	# FIXME: There are various fixes when building under Cygwin
	# or DOS.  These include:
	#   spaces in file names (really too hard/pervasive to fix)
	#   AC_OUTPUT with continuation lines.  Cygwin autoconf
	# doesn't handle these properly.  
	print $FD "\n";
    }
	
}

sub ClearConfigure {
    %ConfigureMakeVars = ();
    %MakeConfigureVars = ();
    %ConfigureDirsUsed = ();
}

# This step handles "compiling" for the Configure step.  In this case, it
# looks for a configure.in file in the current directory and processes it
# to find any dependencies and targets NOT YET IMPLEMENTED
#
# Once it finds the configure, it keeps track of the location in a 
# Stack of configure locations.  This allows us to descend into a submodule
# and then return, keeping track of the correct configure.
# Q: should we have a general register-function to be called on directory
# changes?  This would allow any module to react to directory changes.
sub CompileConfigure {
    if (-s "configure.in") {
	# Temporary
	$gConfigureDir = `pwd`;
	$gConfigureDir =~ s/\r?\n//;
	print "Found a configure.in file\n" if $gDebug;
    }
}

# ----------------------------------------------------------------------------
# Routines exported to other modules
# Clean module calls with (FD, target)
sub ConfigureOutputClean {
    my ($FD, $target) = @_;

    # Do nothing for the clean target
    if ($target eq "distclean") {
	# Remove generated files (e.g., Makefile)
	# FIXME:
	print $FD "\trm -f Makefile\n";
    }
    elsif ($target eq "maintainerclean") {
	# Remove derived files (e.g., configure) and generated files 
	print $FD "\trm -f Makefile\n";
	if (-s "configure.in") {
	    # FIXME: Make this depend on whether a configure was found,
	    # and include any generated headers.  Do *not* include any
	    # Makefiles
	    print $FD "\trm -f configure\n";
	}
    }
}

# Read a configure file and any aclocal.m4 file to understand the configure
$AutoconfDependencies = "";
sub ConfigureAnalyzeConfigure {
    my $aargs = $gAutoconfArgs;
    $AutoconfDependencies = "";

    # Is there an include path in the autoconf args?
    if ($aargs =~ /$gAutoconfIncDirArg\s*([\.\/\w]*)/) {
	my $macroloc = $1;
	# Convert ROOTDIR as necessary
	if ($macroloc =~ /ROOTDIR/) {
	}
	# If we've messed up the location, don't include it
	if (-s "$macroloc/aclocal.m4") {
	    $AutoconfDependencies .= "$macroloc/aclocal.m4";
	    # Extract includes from aclocal.m4
	    open (AFD, "<$macroloc/aclocal.m4" );
	    while (<AFD>) {
		if (/^\s*builtin\(include,([\w-]*\.m4)\)\s*$/) {
		    my $filename = "$macroloc/$1";
		    $AutoconfDependencies .= " $filename";
		}
	    }
	    close AFD;
	}
	else {
	    print STDERR "Could not find expected autoconf macro file $macroloc/aclocal.m4\n";
	}
    }
    
    # Read the configure file for AC_CONFIG_HEADER
}

# Extract derived files from autoconf
# Also record whether there is an AC_CONFIG_HEADER
# If AC_OUTOUT contains any shell variables, you can use
#   <variable_name>_VALUES=vals 
# (all on one line!) to provide the set of possible values.
# Otherwise, no special code will be generated for those output files,
# and a warning message will be issued.
# FIXME: the name is recorded and added to the files, but simplemake
# may not be careful about the file or about directory paths involving the
# xxx_VALUE variable.  Still experimental
# FIXME: clarify which variables are
#   local to ReadAutoconf
#   local to config.pl
#   global
sub ReadAutoconf {
    $configure_has_config_headers = "no";
    %confvars = ();
    open FDCONF, "configure.in" || die "Could not open configure.in file";
    while (<FDCONF>) {
	if (/^\s*AC_CONFIG_HEADER\((.*)\)/) {
	    # Record file name of header in result variable
	    # We also need to add this to the list of files to remove
	    # in the distclean target
	    $configure_has_config_headers = $1;
	}
	if (/^\s*(\w*)_VALUES=(.*)/) {
	    # This is a special case, used below
	    $confvars{$1}=$2
	}
	if (/^\s*AC_OUTPUT\(/) {
	    # First, read past all continued lines
	    $line = "";
	    # Remove any M4 dnl line
	    s/\sdnl\s.*//;
	    while (/\\\s*$/) {
		s/\r?\n//;
		chop($_); # remove the backslash
		$line .= $_;
		$_ = <FDCONF>;
		# Remove any M4 dnl line
		s/\sdnl\s.*//;
	    }
	    $line .= $_; 
	    #chomp( $line );
	    $line =~ s/\r?\n//;
	    $line =~ /^\s*AC_OUTPUT\((.*)\)\s*$/;
	    $files = $1;
	    @autoconf_files = split(/\s+/,$files);
	    # Create a hash by directory
	    $lcurdir = $curdir;
	    $lcurdir =~ s/\/$//;
	    foreach $file (@autoconf_files) {
		if ($file =~ /^(.*)\/([^\/]*)$/) {
		    # File contains a directory path, so update the
		    # appropriate directory entry
		    $autoconf_files_by_dir{"$curdir$1"} .= " $2";
		    #print "{$curdir$1} .= $2\n";
		    # Also add files that are in directories with no simplemake
		    # input (e.g., in include directories with no makefile)
		    if (! -s "$curdir$1/Makefile.sm") {
			$autoconf_files_by_dir{$lcurdir} .= " $file";
		    }
		    else {
			# We also need to handle files that do have a 
			# Makefile.sm,
			# but that aren't in the "usual" path.  Those
			# are directories that are not included in the
			# SUBDIRS path.  These are used only by the distclean
			# target
			$autoconf_files_by_dir_orig{$lcurdir} .= " $file";
		    }
		}
		elsif ($file =~ /^\$(\w*)/) {
		    # Another special case.  The name is actually
		    # a shell variable.  This is too hard for us right now,
		    # so generate a warning and do NOT add it to the list
		    # To handle this case, you can provide
		    # <variable-name>_VALUES
		    $varname = $1;
		    if (defined($confvars{$varname})) {
			# add to autoconf_files_by_dir
			foreach $filename (split(/\s+/,$confvars{$varname})) {
			    $autoconf_files_by_dir{$lcurdir} .= " $file";
			}
		    }
		    else {
			print STDERR "Shell variable $varname will not be added to the list\
of known autoconf file files for ";
			if ($lcurdir ne "") {
			    print STDERR "$lcurdir.\n";
			}
			else {
			    print STDERR "the top directory.\n";
			}
		    }
		}
		elsif ($file =~ /^.*$/) {
		    $autoconf_files_by_dir{"$lcurdir"} .= " $file";
		    #print "{$lcurdir} .= $file\n";
		}
	    }
	    # Add files that configure generates in the current directory
	    $autoconf_files_by_dir{$lcurdir} .= " config.status config.log config.cache *conf.cache config.system";
	    #print "$files\n";
	    last;
	}
    }
    close FDCONF;
}
# This is from simplemake and needs to be updated just like ReadConfigure
#
# Find a working autoconf.
# If the variables autoconf and autoconf_version are set, try to
# find one that works in the user's path.  The format for these
# variables is
#   autoconf = name:name:...:lasthope
#   autoconf_version = number
#   Check each autoconf name for handling version.  If all else fails,
#   unconditionally accept lasthope
sub FindWorkingAutoconf {
    if ($autoconf_version ne "") {
	foreach $file (split(/:/,$autoconf)) {
	    $lastfile = $file;
	    # Grumble.  I need to redirect stderr to stdout before this
	    # open.  We use open with a pipe and then exec so that we
	    # can get stderr redirection to work correctly.
	    #
	    # Additional grumble.  Perl doesn't flush output the
	    # way you'd like.  Of course, if you turn on -debug,
	    # things do get flushed.
	    $pid = open( ACFD, "-|");
	    if ($pid == 0) {
		close FD;  # without flushing
		open STDIN, "/dev/null";
		open STDERR, ">>&STDOUT";
		exec split(/\s+/,"$file --version");
	    }
	    else {
		$found_version = "";
		while (<ACFD>) {
		    print STDERR "$_" if $debug;
		    if (/[Aa]utoconf\s.*([12]\.[0-9]+)/) {
			$found_version = $1;
			last;
		    }
		}
		close( ACFD );
		print STDERR "Autoconf $file is version $found_version\n" if $debug;
		if ($found_version >= $autoconf_version) {
		    $autoconf = $file;
		    print STDERR "Using autoconf $file\n" if $debug;
		    return;
		}
	    }
	}
	# If we got here, use the lastchance
	$autoconf = $lastfile;
    }
}

1;
