#
# Provide routine to convert a C program from text to HTML.
# DOC
# CToHtml - Convert a C program on an input stream into HTML on the output
#           stream
#
# ENDDOC
#
%CKeywords = ( int => 1,
	       double => 1, 
	       short => 1,
	       unsigned => 1,
	       signed => 1,
	       float => 1,
	       char   => 1, 
	       long   => 1,
	       void   => 1,
	       wchar_t => 1,

	       const => 2,
	       extern => 2,
	       register => 2,
	       volatile => 2,
	       static => 2,
 	       auto => 2,
	       restrict => 2,
	       
	       typedef => 3,
	       union => 3,
	       enum => 3,
	       struct => 3,
	       if => 3,
	       else => 3,
	       case => 3,
	       for => 3,
	       while => 3,
	       do => 3,
	       switch => 3,
	       return => 3,
	       goto => 3,
	       break => 3,
	       continue => 3,
	       default => 3,

	       sizeof => 4,

	       # Library routines and system calls (this is a partial list,
	       # but it matches the one from c2html)
	       abort => 5,
	       abs => 5,
	       acos => 5,
	       asctime => 5,
	       asin => 5,
	       atan => 5,
	       atan2 => 5,
	       atexit => 5,
	       atof => 5,
	       atoi => 5,
	       atol => 5,
	       bsearch => 5,
	       calloc => 5,
	       ceil => 5,
	       clearerr => 5,
	       clock => 5,
	       cos => 5,
	       cosh => 5,
	       ctime => 5,
	       difftime => 5,
	       div => 5,
	       exit => 5,
	       exp => 5,
	       fabs => 5,
	       fclose => 5,
	       feof => 5,
	       ferror => 5,
	       fflush => 5,
	       fgetc => 5,
	       fgetpos => 5,
	       fgets => 5,
	       floor => 5,
	       fmod => 5,
	       fopen => 5,
	       fprintf => 5,
	       fputc => 5,
	       fputs => 5,
	       fread => 5,
	       free => 5,
	       freopen => 5,
	       frexp => 5,
	       fscanf => 5,
	       fseek => 5,
	       fsetpos => 5,
	       ftell => 5,
	       fwrite => 5,
	       getc => 5,
	       getchar => 5,
	       getenv => 5,
	       gets => 5,
	       gmtime => 5,
	       isalnum => 5,
	       isalpha => 5,
	       iscntrl => 5,
	       isdigit => 5,
	       isgraph => 5,
	       islower => 5,
	       isprint => 5,
	       ispunct => 5,
	       isspace => 5,
	       isupper => 5,
	       isxdigit => 5,
	       labs => 5,
	       ldexp => 5,
	       ldiv => 5,
	       localtime => 5,
	       log => 5,
	       log10 => 5,
	       longjmp => 5,
	       malloc => 5,
	       memchr => 5,
	       memcmp => 5, memcpy => 5, memmove => 5, memset => 5,
	       mktime => 5, modf => 5, perror => 5, pow => 5, printf => 5, 
	       putc => 5, putchar => 5, puts => 5, qsort => 5, raise => 5, 
	       rand => 5, realloc => 5, remove => 5, rename => 5, rewind => 5, 
	       scanf => 5, setbuf => 5, setvbuf => 5, setjmp => 5, 
	       signal => 5, sin => 5, sinh => 5, sprintf => 5, sqrt => 5, 
	       srand => 5, sscanf => 5, strcat => 5, strchr => 5, strcmp => 5, 
	       strcpy => 5, strcspn => 5, strerror => 5, strftime => 5, 
	       strlen => 5, strncat => 5, strncmp => 5, strncpy => 5, 
	       strpbrk => 5, strrchr => 5, strspn => 5, strstr => 5, 
	       strtod => 5, strtok => 5, strtol => 5, strtul => 5, 
	       system => 5, tan => 5, tanh => 5, time => 5, tmpfile => 5, 
	       tmpnam => 5, tolower => 5, toupper => 5, ungetc => 5, 
	       vfprintf => 5, vprintf => 5, vsprintf => 5, 
	       close => 5, create => 5, fstat => 5, lseek => 5, open => 5, 
	       read => 5, 
	       abrk => 5, stat => 5, unlink => 5, write=>5,

	       );
%ColorClass = ( 1 => "4169E1", 
		2 => "4169E1",
		3 => "4169E1",
	        4 => "4169E1",
		5 => "4169E1",
		);
# nClasses is the number of classes in use; the classes after this are
# available for user-defined extensions
$nClasses = 5;
$commentColor = "B22222";
$cppColor     = "A020F0";
$stringColor  = "666666";
$includeColor = "A020F0";                # Special color for #include
$defineColor  = "228B22";
$newline      = "\r\n";                  # this newline is better for Windows

# Function definition <strong>?  Also includes name block with the function's
# name

#
# CToHtml( readfunc, readdata, writefunc, writedata )
sub CToHtml {
    my ($readFunc, $readData, $writeFunc, $writeData) = @_;

    &$writeFunc( $writeData, "<pre width=\"80\">$newline" ) || return 0;
    my $atLeft    = 1;
    while (1) {
	($_,$hasErr) = &$readFunc( $readData );
	if ($hasErr) { last; }

	$atLeft = 1;
	s/\r?\n//;
	print STDOUT "Read $_\n" if $gDebug;
	my $loopcount = 0;  # Used to intercept infinite loops
	while (/^(\s*)(\w*)(.*)/ && $loopcount++ < 100) {
	    my $spaces = $1;
	    my $token  = $2;
	    my $after  = $3;

	    if ($gDebug) {
		if ($token ne "") {
		    print STDOUT "\tToken $token\n";
		}
		else {
		    my $first = substr $after, 0, 1;
		    print STDOUT "\tNext token not a word ($first)\n";
		}
	    }
	    
	    &$writeFunc( $writeData, $spaces ) || return 0;
	    
	    if ($token ne "") {
		# There is an alphanumeric token to look at
		if (defined($CKeywords{$token})) {
		    $keycolor = $ColorClass{$CKeywords{$token}};
		    &$writeFunc( $writeData, 
				   "<font color=\"$keycolor\">" . $token . 
				   "</font>" ) || return 0;
		}
		else {
		    &$writeFunc( $writeData, $token ) || return 0;
		}
	    }
	    else {
		# Handle non-word characters
		$_ = $after;
		# ? should never have spaces at the beginning 
		# We need to match a non-space, non-word character
		if (/^(\s*)(\W+)(.*)/) {
		    $spaces = $1;
		    $token  = $2;
		    $after  = $3;

		    # Since \W will match spaces, we need to adjust it
		    if ($token =~ /(\S*)(.*)/) {
			$token = $1;
			$after = $2 . $after;
		    }

		    print STDOUT "\tnon-word token :$token:\n" if $gDebug;
		    &$writeFunc( $writeData, $spaces ) || return 0;
		    # Check for special cases:
		    # /* ... */ comment 
		    # ^#\s*command (CPP) (if atLeft)
		    # strings, character constants ("..." and 'c')
		    if ($token =~ /^\/\*(.*)/) {
			$after = $1 . $after;
			($comment, $after) = 
			    &ReadComment( $readFunc, $readData, $after );
			&$writeFunc( $writeData, 
				       "<font color=\"$commentColor\">/*" .
				       &HTMLify( $comment ) . 
				       "</font>" ) ||
					   return 0;
		    }
		    elsif ($token =~ /\/\//) {
			# C++ comment.
			&$writeFunc( $writeData, 
				       "<font color=\"$commentColor\">//" . 
				       &HTMLify( $after ) . "</font>" ) ||
			    return 0;
			$after = "";
		    }
		    elsif ($token =~ /^\"(.*)/) {
			# a string.  Match the end of the string
			# Make after the entire input
			$after = $1 . $after;
			($string, $after) = 
			    &ReadString( $readFunc, $readData, $after );
			&$writeFunc( $writeData, 
				       "<font color=\"$stringColor\">\"" .
				       &HTMLify( $string ) . 
				       "</font>" ) || return 0;
		    }
		    # Should we do something for single characters (e.g., 
		    # 'c')?
		    elsif ($atLeft && $token eq "#") {
			# Handle the subcases:
			#   include
			#   define
			# other
			$after =~ /^(\s*)(\w*)(.*)/;
			$spaces = $1;
			$token  = $2;
			$after  = $3;
			print STDOUT "Cpp token is $token\n" if $gDebug;
			if ($token eq "define") {
			    &$writeFunc( $writeData, 
					   "<font color=\"$defineColor\">#" .
					   $spaces . $token . "</font>" . 
					   &HTMLify($after) ) ||
				return 0;;
			}
			elsif ($token eq "include") {
			    &$writeFunc( $writeData, 
					   "<font color=\"$includeColor\">#" .
					   $spaces . $token . "</font>" .
					   &HTMLify($after) ) ||
				return 0;
			}
			else {
			    &$writeFunc( $writeData, 
					   "<font color=\"$cppColor\">#" .
					   $spaces . &HTMLify( $token ) . 
					   "</font>" . &HTMLify( $after ) ) ||
					   return 0;
			}
			$after = "";
		    }
		    else {
			# Carve off the first character (to avoid matching
			# the multiple characters ', "' (comma,space,double-quote)
			if ($token =~ /^(.)(\s*)(.*)/) {
			    $token  = $1;
			    $spaces = $2;
			    $after  = $3 . $after;
			    # HTMLify the token
			    &$writeFunc( $writeData, &HTMLify( $token ) .
					   $spaces ) || return 0;
			}
			# Else token is empty and doesn't match at all
		    }
		}
	    }
	    $_ = $after;
	    if (! /\S/) { last; }
	    $atLeft = 0;
	}
	if ($loopcount >= 100) {
	    print STDERR "PANIC: problem in while\n";
	}
	&$writeFunc( $writeData, $newline ) || return 0;
    }
    &$writeFunc( $writeData, "</pre>$newline" ) || return 0;

    # Return standard zero on success
    return 1;
}

# -------------------------------------------------------------------------
# Need to move these into C support
# -------------------------------------------------------------------------
# Read a comment.  The current line is in $after; $token is empty
# Return value is the "leftover" (after) characters
sub ReadComment {
    my ($readFunc, $readData, $after) = @_;
    my $token, $comment, $isErr;
    
    $comment = "";
    $isErr = 0;
    # comment.  Read until matching end-comment
    # Read until the end of string is found
    while (! $isErr) {
	print STDOUT "Reading comment...\n" if $gDebug;
	if ($after =~ /(.*\*\/)(.*)/) {
	    $token = $1;
	    $after = $2;
	    $comment .= $token;
	    last;
	}    
	$comment .= $after;
	($after,$isErr) = &$readFunc( $readData );
	$after =~ s/\r?\n//;
    }
    return ($comment, $after);
}

sub ReadString {
    my ($readFunc, $readData, $after) = @_;
    my $token, $string;

    $string = "";
    
    print STDOUT "Reading string\n" if $gDebug;
    while ($after =~ /^([^\"]*)(.*)/) {
	$token = $1;
	$after = $2;
	print STDOUT "\ttoken $token\n" if $gDebug;
	$string .= $token;
	if ( ($after =~ /^\"/) && ($token =~ /\\$/) ) {
	    # Keep going; strip the quote out of after
	    $string .= "\"";
	    $after = substr $after,1;
	}
	else {
	    if ($after =~ /^\"/) {
		$string .= "\"";
		$after =~ s/^\"//;
		last;
	    }
	    # The other case is $after is empty
	}
	if (! ($after =~ /\S/) ) {
	    ($line,$isErr) = &$readFunc( $readData );
	    $after .= $line;
	}
    }
    print STDOUT "Done reading string\n" if $gDebug;
    return ($string,$after);
}

