#
# This module handles the documentation targets
# These are subsets of the SOURCES targets; i.e.,
# html_DOC_SOURCES = 
# latex_DOC_SOURCES = 
#
# We have the following issues to be concerned with:
#   Suffixes (e.g., .html, .man, .tex)
#   Install 
#      program
#      directory
#   
# TODO: This is a subset of the SOURCES action
# Note that this does NOT register itself as a known action, since 
# it is a subcase of the SOURCES action
#
# To do: subinstall routines
#
# Exported:
#    SOURCE_OutputHeader_SOURCE_DOC - invoked when DOC_SOURCE is seen by
# the Source module
#
# For differnent types of documents, this hash gives the installation 
# directories
%DOCInstallDirs = ( "html" => "www/www3", 
		    "man" => "man/man3", 
		    "latex" => "doc/refman" );

# man targets for documentation, based on the type
%DOCTargetName = ( "html" => "htmldoc",
		   "man" => "mandoc",
		  "latex" => "latexdoc" );
# If a "DOC" target is seen at the top file, remember this.  This will 
# ensure that DOC targets are generated in the appropriate directories
$gDOCAtTop = 0;

# docFiles provides a mapping from a doc type to a list of files
%docFiles = ();

# Should these be SOURCE_Action_SOURCE_DOC etc?
sub ActionDOC {
    my ($sub,$term,$value,$filename,$linecount) = @_;

    if ($value =~ /\S/) {
	# Ignore empty values
	$docFiles{$sub} .= " $value";
    }
}

sub ClearDOC {
    %docFiles = ();
}

sub OutputDOC {
    my $FD = $_[0];

    
}

sub SOURCES_OutputHeader_SOURCE_DOC {
    my $FD = $_[0];
    
    print $FD "doctext       = \@doctext\@\n";
    # Let everyone know that we're defining a doctext variable
    &ConfigureRegister( "doctext", "doctext" );
}

# Variables used to control the output formats
# FIXME: How do these get set (in conf file?)
$gDOCHeading = "";
$gDOCDefsFiles = "";

sub SOURCES_OutputDefault_SOURCE_DOC {
    my $FD = $_[0];
    
    # for each source definition that I've seen, generate the appropriate
    # default target 

    # Currently, we only know about .c (and .h) files
    if (defined($docFiles{"man"})) {
	print $FD ".c.man:\n";
	print $FD "\t\$(doctext) -mpath $dest -ext 3 -heading $gDOCHeading \\\n";
	print $FD "\t\t -quotefmt $gDOCDefsFiles \$<\n";
	&ConfigureRegister( "mandir", "\@mandir\@" );
    }
    if (defined($docFiles{"html"})) {
	print $FD ".c.html:\n";
	print $FD "\t\$(doctext) -html -mpath $dest -heading $gDOCHeading \\\n";
	print $FD "\t\t -quotefmt $gDOCDefsFiles \$<\n";
	&ConfigureRegister( "htmldir", "\@htmldir\@" );
    }
    if (defined($docFiles{"tex"})) {
	print $FD ".c.tex:\n";
	print $FD "\t\$(doctext) -latex -mpath $dest -heading $gDOCHeading \\\n";
	print $FD "\t\t -quotefmt $gDOCDefsFiles \$<\n";
	&ConfigureRegister( "docdir", "\@docdir\@" );
    }

    &SUBDIRSRecursiveOp( $FD, "doc" );

    
}
# This is not a known action, it is a subaction for SOURCE.
#$KnownActions[$#KnownActions] = "DOC";

1;
