#
# Install targets
#
# Example Makefile Use:
# LIB_INSTALL = libfoo.a 
# BIN_INSTALL = p1 prog2 
# BIN_OPTINSTALL = otherprog
# 
$KnownActions[$#KnownActions+1] = "INSTALL";
$KnownActions[$#KnownActions+1] = "OPTINSTALL";

# Local data for the install operations
#install_dir maps XXX from LIB_INSTALL into the GNU name
%install_dir = ( 'LIB' => 'libdir', 
		 'SHLIB' => 'libdir',     # SHLIB also signals library to build
		 'BIN' => 'bindir',
		 'INCLUDE' => 'includedir', 
		 'MAN' => 'mandir',
		 'HTML' => 'htmldir',
		 'DOC' => 'docdir',
		 'ETC' => 'sysconfdir',
		);
# We add the -p switch to install for the libraries because libraries
# that require ranlib may fail if installed without preserving the
# data/time on the file.  This should work because configure looks for
# a BSD-style install.
# Sigh.  ginstall (!!!) doesn't support the -p switch.
# This means that there is no way to install a library under Darwin
# without defining INSTALL_DATA as install -p, which configure 
# does not figure out(!)
%install_methods = ( 'LIB'     => '$(INSTALL_DATA)',
		     'BIN'     => '$(INSTALL_PROGRAM) $(INSTALL_STRIP_FLAG)',
		     'SHLIB'   => '$(INSTALL_PROGRAM)',
		     'INCLUDE' => '$(INSTALL_DATA)',
		     'MAN'     => '$(INSTALL_DATA)',
		     'HTML'    => '$(INSTALL_DATA)',
		     'DOC'     => '$(INSTALL_DATA)',
		     'ETC'     => '$(INSTALL_DATA)',
		    );
# These give the directories that autoconf will require for the 
%required_dirs = ( 'libdir'      => 'exec_prefix prefix',
		   'bindir'      => 'exec_prefix prefix',
		   'mandir'      => 'prefix',
		   'htmldir'     => 'prefix',
		   'docdir'      => 'prefix',
		   'includedir'  => 'prefix',
 		   'exec_prefix' => 'prefix',
		   'sysconfdir'  => 'prefix',
		  );

# Install files is a hash that maps the kind to a list of files to be installed
%installFiles = ();
%optinstallFiles = ();
sub ActionINSTALL {
    my ($sub,$term,$value,$filename,$linecount) = @_;

    # Using subname, determine the appropriate operations
    if (!defined($install_methods{$sub})) {
	print STDERR "Unknown install case $sub\n";
	return 0;
    }
    &ConfigureRegister( "INSTALL", "\@INSTALL\@" );
    &ConfigureRegister( "INSTALL_DATA", "\@INSTALL_DATA\@" );
    &ConfigureRegister( "INSTALL_PROGRAM", "\@INSTALL_PROGRAM\@" );
    # FIXME: how do we ensure that there is a need for and a value provided
    # for MKDIR_P?
    &ConfigureRegister( "MKDIR_P", "\@MKDIR_P\@" );
    
    $installFiles{$sub} .= " $value";
}
sub ActionOPTINSTALL {
    my ($sub,$term,$value,$filename,$linecount) = @_;

    # Using subname, determine the appropriate operations
    if (!defined($install_methods{$sub})) {
	print STDERR "Unknown optinstall case $sub\n";
	return 0;
    }
    &ConfigureRegister( "INSTALL", "\@INSTALL\@" );
    &ConfigureRegister( "INSTALL_DATA", "\@INSTALL_DATA\@" );
    &ConfigureRegister( "INSTALL_PROGRAM", "\@INSTALL_PROGRAM\@" );
    # FIXME: how do we ensure that there is a need for and a value provided
    # for MKDIR_P?
    &ConfigureRegister( "MKDIR_P", "\@MKDIR_P\@" );
    
    $optinstallFiles{$sub} .= " $value";
}

# These is no special header since all we need is the configure variables (?)
sub OutputHeaderINSTALL {
}

# Note that there is no separate OPTINSTALL output routine; instead, 
# the install target includes the optional files, using a slightly 
# different commandline
sub OutputINSTALL {
    my $FD = $_[0];
    my %dirschecked = ();  # A hash of directories that are in use.
    
    # See if there is anything to do; return immediately if not
    if (!scalar(%installFiles) && !scalar(%optinstallFiles)) { 
	# Only return if there are no subdirectories
	# We import the array from SUBDIRS that remembers the
	# install directories, if any
	if (!defined(@SUBDIRS_INSTALL) || $#SUBDIRS_INSTALL == -1) {
	    print "No files to install\n" if $gDebug;
	    return; 
	}
    }
    
    print $FD "install: FORCE_TARGET\n";

    # First, create any directories needed

    my $lastkind = "";
    foreach my $kind (sort( keys(%installFiles), keys(%optinstallFiles))) {
	# skip duplicates
	if ($kind eq $lastkind) { next; }
	$lastkind = $kind;
	# $dir is actually the name of a make variable
	my $dir = $install_dir{$kind};
	if ($dir eq "") {
	    print STDERR "No known installation dir for install_$kind\n";
	    next;
	}

	# The first step is to make sure that all of the necessary 
	# directories exist
	# FIXME: 
	#   We need to register MKDIR_P
	print $FD "\tif [ ! -d \$($dir) ] ; then \$(MKDIR_P) \$($dir) ; fi\n";
    }	

    foreach my $kind (sort(keys(%installFiles))) {
	# Then, for each for the files, we install it with the proper method
	my $dir = $install_dir{$kind};
	foreach my $file (split(/\s+/,$installFiles{$kind})) {
	    if (! ($file =~ /\S/)) { next; }
	    my $destfile = $file;
	    # Remove any directories from the destination file name
	    $destfile =~ s/.*\///;
	    $this_install_method = $install_methods{$kind};
	    if ($this_install_method eq "") {
		# Use the default method
		$this_install_method = '$(INSTALL)';
		}
	    print $FD "\t$this_install_method $file \${$dir}/$destfile\n";
	}

    }

    foreach my $kind (sort(keys(%optinstallFiles))) {
	# Then, for each for the files, we install it with the proper method
	foreach my $file (split(/\s+/,$optinstallFiles{$kind})) {
	    if (! ($file =~ /\S/)) { next; }
	    my $destfile = $file;
	    # Remove any directories from the destination file name
	    $destfile =~ s/.*\///;
	    $this_install_method = $install_methods{$kind};
	    if ($this_install_method eq "") {
		# Use the default method
		$this_install_method = '$(INSTALL)';
		}
	    print $FD "\tif [ -s $file ] ; then $this_install_method $file \${$dir}/$destfile ; fi\n";
	}

    }
    # invoke any subdir installs
    &SUBDIRSRecursiveOp( $FD, "install", "INSTALL" );

    print $FD "\n";
}

# Register the directories that we need with the configure module
sub CompileINSTALL {
    $lastkind = NONE;
    foreach my $kind (sort( keys(%installFiles), keys(%optinstallFiles))) {
	# skip duplicates
	if ($kind eq $lastkind) { next; }
	$lastkind = $kind;
	# $dir is actually the name of a make variable
	my $dir = $install_dir{$kind};
	if ($dir ne "") {
	    &ConfigureRegisterDirUsed( $dir );
	}
    }	
}

sub ClearINSTALL {
    %installFiles = ();
    %optinstallFiles = ();
}
1;
