# -*- Mode:Perl; -*-
# ---------------------------------------------------------------------------
# RunProgram directory LIST
# 
# Output is passed through routines 
#    $OutputLine
# with $OutputClose, $OutputReopen used to close files used for execed process
# This routine returns the return code and signal num (signal 0 means no 
# signal)
#
# FIXME:
#    Detect shell meta characters (e.g., >, >>) in command.  Even better
#    would be to handle them.
sub RunProgram {
    my ($rundir, @args) = @_;
    my $signalNum = 0;
    my $rc = 0;

    # Validate the inputs first
    if (!defined($args[0]) || $args[0] eq "") {
	print STDERR "No program given to RunProgram\n";
	return (1,"","");
    }

    # perl does not correctly handle "2>&1 >>foo" redirection in system
    # correctly (some stderr escapes to the prior stderr).  This
    # code attempts to do the correct thing.
    # By reopening stdout and stderr, we can ensure that all of the 
    # output goes to the specified files.  Unfortunately, we can't
    # force perl to correctly flush files without open/close, so
    # we close the output file before the fork.  This guarantees that
    # the data is flushed.  We reopen it after the fork (which is implicit
    # in the open)
    &$OutputClose();

    # We now use a different approach that uses a open that creates a 
    # fork and associates file handle (CFD for Child FD).
    my $pid = open(CFD,"-|");
    if ($pid == 0) {
	# we're the child
        open STDIN, '/dev/null';
	# FIXME: Do we want to allow an output filter, e.g., to convert the 
	# output into well-formed XML?
        open STDERR, ">>&STDOUT";
	if (defined($rundir) && $rundir ne "") {
	    if (-d $rundir) {
		$rc = chdir $rundir;
		if ($rc == 0) {
		    return ($rc,0,"Could not change to $rundir");
		}
	    }
	    else {
		return (1,0,"Could not find $rundir");
	    }
	}
	#for (my $i=0; $i<=$#args; $i++) {
	#    print "Arg $i: $args[$i]\n";
	#}
	# See perlfunc for details on the following construction.  This
	# keeps perl from handling arrays with one element different from
	# those with more elements
        exec { $args[0] } @args;
	die "Could not exec program $args[0]\n";
    }
    else {
	# Read from the child
	&$OutputReopen();
	while (<CFD>) {
	    &$OutputLine( $_ );
	}
	# Closing a pipe implicitly waits for the command on the other
	# end to complete, and puts the exit status into $?
	close CFD;
	# end of read from the child
	# Note that this status is usually shifted right by 8, so
	# we check for that
        $rc = $?;
	$signalNum = $rc & 127;
	print STDERR "RunProgram rc = $rc and sig = $signalNum\n" if $gDebug;
	if ($rc > 255) { $rc = $rc >> 8; }
    }
    return ($rc,$signalNum,"");
}

#
# This routine is a convenience function to report on the status of
# a run program; it takes that output from RunProgram
sub ProgramReport {
    my ( $rc, $signal, $msg ) = @_;
    if ($rc == 0 && $signal == 0) { return 0; }
    if ($signal) {
	print STDERR "Program exited with signal $signal\n";
    }
    if ($msg ne "") {
	print STDERR "RunProgram returned message $msg\n";
    }
    return $rc;
}

1;
