# ---------------------------------------------------------------------------
# We need to split this into subactions
# (we need a more consistent naming scheme)
# SOURCES naming scheme:
#
#  name_kind_SOURCES
#  Special case:
#     name_SOURCES - for simple programs
#  name_LIB_SOURCES
#  (todo:)
#  name_SHLIB_SOURCES
#  name_DOC_SOURCES (e.g., HTML_DOC_SOURCES)
#  name_PGM_SOURCES to define a program with an embedded _ in it.
#
# A number of steps (e.g., TAGS, DEPENDENCIES) depend on the source files.
# The following variables make the source files available to other modules
#  AllSources{ext}
# e.g., AllSources{c} gives all of the C sources (particularly valuable 
# for TAGS)
# <type>Sources{name} gives the sources for things of <type> and the
# given name; e.g., LibSources{foo} gives the sources for libfoo and
# Sources{bar} gives the sources for program bar.
#
# todo: What about documentation targets, like pdf.  
# Should we include paper_PDF_SOURCES ?
# Library dependencies for program? (see simplemake)
#
# todo: Add EXECS variable, (EXTRA_EXECS); used for clean
#       Add C_SOURCES, C_HEADERS, F77_SOURCES, TEX_SOURCES, etc. variables; 
#        used for TAGS
#
# Subactions is an array of known subcases, such as LIB and PGM
@SOURCESubactions = ();
# Extensions is a list of extensions seen in this input
%SOURCEExtensions = ();
# Target extensions is a list of extensions that are targets for the
# sources seen (e.g., a .c extension adds o to TargetExtensions, 
# .tex adds dvi and pdf).
%SOURCETargetExtensions = ();
#
# %SOURCE_subaction{name} => value is the sources for each name for the
# target name and the specified subaction (e.g., PGM or LIB)
# 
sub ActionSOURCES {
    my ($sub,$term,$value,$filename,$linecount) = @_;

    &ConfigureRegister( "srcdir", "\@srcdir\@" );
    # When do we need these next two variables
    &ConfigureRegister( "top_srcdir", "\@top_srcdir\@" );
    &ConfigureRegister( "master_top_srcdir", "\@master_top_srcdir\@" );

    # Does sub contain a known subcase?
    if ($sub =~ /^(\S*)_([^_\s]*)/) {
	my $name      = $1;
	my $subaction = $2;
	print "Found $subaction for $name\n" if $gDebug;
	$hashname = "SOURCE_$subaction";
	if (!defined(%$hashname)) {
	    $SOURCESubactions[$#SOURCESubactions+1] = $hashname;
	}
	$$hashname{$name} = $value;
	# FIXME: need to register sources here.
	# Problem: need to adjust for subtype; must invoke function?
	my $funcname = "SOURCE_${subaction}TrackSources";
	if (defined(&$funcname)) {
	    print "Invoking $funcname($sub,$name,$value)\n" if $gDebug;
	    &$funcname( $name, $value );
	}
	#&SOURCE_Register_<subname>( name )
	my $func = "SOURCE_${subaction}ConfigureVars";
	if (defined(&$func)) {
	    &$func;
	}
    }
    else {
	# Simple program case (alias for PGM)
	# Remember that we've seen a program
	$hashname = "SOURCE_PGM";
	if (!defined(%$hashname)) {
	    $SOURCESubactions[$#SOURCESubactions+1] = $hashname;
	}
	# Get the name of the program from the subterm.
	my $pgm = $sub;
	$pgm =~ s/_$//;
	print "Found program $pgm\n" if $gDebug;

	$$hashname{$pgm} = $value;
	# Split value into individual items and track the extensions
	&SOURCETrackExtensions( $value );
	# Add "no-extension" as a target.
	$SOURCETargetExtensions{BLANK} = 1;
	$SOURCETargetExtensions{o}     = 1;

	# FIXME: We need to register after reading the file in case
	# a target is an "extra" instead of an "all" target.
	&SMRegisterForAll( $pgm );
    }
}
sub ClearSOURCES {
    foreach my $hashname (@SOURCESubactions) {
	# Activate the clear function for the subaction
	my $function = "SOURCES_Clear_SOURCE_$hashname";
	if (defined(&$function)) {
	    &$function();
	}
		
	%$hashname = ();
	undef(%$hashname);
    }
    @SOURCESubactions = ();
    %SOURCEExtensions = ();
    %SOURCETargetExtensions = ();
}

sub OutputSOURCES {
    my $FD = $_[0];
    
    # Generic rules first
    foreach my $ext (keys(%SOURCEExtensions)) {
	# For each extension, we need to know the chain of dependencies
	# that we will use.  These depend on the source type,
	# so we will need to be careful here
	# FIXME: Keep track of which rules are needed 
	# 
	foreach my $targetExt (keys(%SOURCETargetExtensions)) {
	    # BLANK is a special case for no extension on the 
	    # target; used for programs
	    my $tsep = ".";
	    if ($targetExt =~ /BLANK/) {
		$targetExt = "";
		$tsep      = "";
	    }
	    my $rule = $SOURCEBuildRule{"$ext:$targetExt"};
	    if ($rule ne "") {
		print $FD ".$ext$tsep$targetExt:\n";
		print $FD "\t" . $rule . "\n";
	    }
	}
    }
    print $FD "\n";

    foreach my $hashname (@SOURCESubactions) {
	print "need $hashname\n";
	my $function = "SOURCES_Output_$hashname";
	if (defined(&$function)) {
	    &$function( $FD, $hashname );
	}
    }
    foreach my $lib (keys(%SourceLibraries)) {
	;
    }
    foreach my $pgm (keys(%SOURCE_PGM)) {
	my $value = $SOURCE_PGM{$pgm};
	print $FD "$pgm: ";
	# Print the rule that builds this program.  Use the following
	# rules:
	#    If all sources have a single extension (or .o)
	# use the rule .ext: 
        # 
	my $ext = &SOURCEFindExtensions( $value );
	if ($ext ne "") {
	    print $FD "$value\n";
	    my $rule = $SOURCEBuildRule{"$ext:"};
	    print $FD "\t$rule\n";
	}
	else {
	    # Replace value with .o files, then use the link rule
	    my $dvalue = $value;
	    # FIXME: Not quite right
	    $dvalue =~ s/\.\w*/.o/g;
	    print $FD "$dvalue\n";
	    print $FD "\t$linkrule $dvalue\n";
	}
    }
}
sub OutputHeaderSOURCES {
    my $FD = $_[0];
    
    foreach my $hashname (@SOURCESubactions) {
	print "need $hashname\n";
	my $function = "SOURCES_OutputHeader_$hashname";
	if (defined(&$function)) {
	    &$function( $FD, $hashname );
	}
    }
    # Need header info for Lib and Programs
    # FIXME:
    # if there are C/C++ programs, we need
    # DEFS = <otherdefs> @DEFS@
    # INCLUDES = <includelist>
    # CPPFLAGS = <otherflags> @CPPFLAGS@
    # LIBS = <otherlibs> @LIBS@
    
    # For each source extension, see if we 
    # have a default rule
    foreach my $ext (keys(%SOURCEExtensions)) {
	# For each extension, we need to know the chain of dependencies
	# that we will use.  These depend on the source type,
	# so we will need to be careful here
	# FIXME: Keep track of which rules are needed 
	# 
	foreach my $targetExt (keys(%SOURCETargetExtensions)) {
	    # BLANK is a special case for no extension on the 
	    # target; used for programs
	    $targetExt =~ s/BLANK//;
	    my $rule = $SOURCEHeaderRule{"$ext:$targetExt"};
	    if ($rule && $rule ne "") {
		print $FD $rule . "\n";
	    }
	}
    }
    
}
sub OutputDefaultSOURCES {
    my $FD = $_[0];
    
    foreach my $hashname (@SOURCESubactions) {
	print "need $hashname\n";
	my $function = "SOURCES_OutputDefault_$hashname";
	if (defined(&$function)) {
	    &$function( $FD, $hashname );
	}
    }
    # Need header info for Programs
}

# ----------------------------------------------------------------------------
# LIB_SOURCES
#
# Make depends on file timestamps to determine some dependencies.  
# On systems with networked file systems, the time stamp for the local disk
# may not be consistent with the file system, leading to problems with
# determining which files should be updated.  This code attempts to fix that,
# in the fast case by resetting the time stamp (using a short PERL program)
# or in the fast case by simply sleeping a short time (1 second in the
# code from simplemake).
#
# ---- from simplemake
$fixup_for_timestamps = 1; # make is broken
# FIXME:
# The sleep amount should be parameterized.  One developer at IBM 
# needed 10 seconds, not 1, because of the variation in file system times
# between the local machine and the file server.  Perhaps something like
# \@INTERDIR_SLEEP\@ or 
# if [ -z "$(INTERDIR_SLEEP)" ] ; then INTERDIR_SLEEP=1 ; fi ; sleep $(INTERDIR_SLEEP)
$SleepSlow = "sleep 1";
$SleepFast = "\@PERL\@ -e \'\$\$now=time;\$\$now-=10;utime \$\$now, \$\$now, \"LIB\";\'";
$Sleep = $SleepFast;
# ---- end from simplemake

sub SOURCES_Output_SOURCE_LIB {
    my $FD       = $_[0];
    my $hashname = $_[1];

    print "Looking at $hashname\n" if $gDebug;
    foreach my $lib (keys(%$hashname)) {
	print "Found $lib\n" if $gDebug;
	my $sources = $$hashname{$lib};
	my $objects = &SOURCESToObjects( $sources );
	my $libloc  = &SOURCE_LibGetLibLocation( $lib );
	print $FD "lib$lib.a: $objects\n";
	print $FD "\t\$(AR) cr lib$lib.a \$?\n";
	print $FD "\t\$(RANLIB) lib$lib.a\n";
	# FIXME: Here goes whatever is necessary to keep make from
	# making a mistake for subsequent changes to this library
	# (either sleep $gSyncTime or the perl code to change
	# the modification time
    }
}
# Each subtype can register the configure variables that it may need
sub SOURCE_LIBConfigureVars {
    &ConfigureRegister( "AR", "\@AR\@" );
    &ConfigureRegister( "RANLIB", "\@RANLIB\@" );
}

#
# FIXME: We need a way to place the library in a different directory.
# 
# Pass the name of the library to this routine, which returns
# a path to the library
sub SOURCE_LibGetLibLocation {
    my $libname = $_[0];

    return "";
}

# (libname,value)
sub SOURCE_LIBTrackSources {
    my ($name,$value) = @_;
    &SOURCETrackExtensions( $value );
    $SOURCETargetExtensions{o}     = 1;
}

sub CompileSOURCE {
}
# ----------------------------------------------------------------------------
# This rourine tracks source extensions that have been seen.  The input
# is a space-separated list of files
sub SOURCETrackExtensions {
    my @filelist = split(/\s+/,$_[0]);

    foreach my $file (@filelist) {
	my $ext = "";
	if ($file =~ /(.*)\.(\w*)$/) {
	    $ext = $2;
	    $SOURCEExtensions{$ext} = 1;
	}
    }
}

# return a space-separated list of the generic extensions (those
# for which this module will provide default rules)
sub SOURCESGetSuffixes {
    my @suffixes = sort(keys(%SOURCEExtensions),keys(%SOURCETargetExtensions));
    my $list = "";
    my $lastname = "";
    for (my $i=0; $i<=$#suffixes; $i++) {
	my $name = $suffixes[$i];
	if ($name eq $lastname) { next; }
	if ($name eq "BLANK") { next; }
	$list .= "$name ";
	$lastname = $name;
    }
    $list =~ s/ $//;  # remove trailing blank, if any.
    return $list;
}

sub SOURCEOutputClean {
    my ($FD,$target) = @_;

    # For programs, we need to clean .o
    # For libraries, we also need to clean .o 
    #    (these two may want to communicate that the other has removed 
    #    .o files)
    # For LaTeX documents, we need to clean .dvi, .aux, .log
    # For configure, also .log
    # There are also some common fils to clean, such as *~

    # Handle these cases by invoking sub action methods
    foreach my $hashname (@SOURCESubactions) {
	my $function = "SOURCES_OutputClean_$hashname";
	print "Looking for $function\n" if $gDebug;
	if (defined(&$function)) {
	    &$function( $FD, $target );
	}
    }
}

# ---------------------------------------------------------------------------
# Programs
sub SOURCE_OutputClean_PGM {
    my ($FD,$target) = @_;
    
    print $FD "\trm -f *.o\n";
}
# ---------------------------------------------------------------------------
$KnownActions[$#KnownActions+1] = "SOURCES";

# Default definitions for source files
# FIXME: Also need to separate out .c -> .o and .c -> .lo
# Most generally, for a given source extension and destination type,
# need to specify source and destination extensions.

%SOURCEHeaderRule = (
		     'c:o' => 
"CC          = \@CC\@
CFLAGS      = \@CFLAGS\@
C_COMPILE   = \$(CC) \$(DEFS) \$(INCLUDES) \$(CFLAGS) \$(CPPFLAGS)", 
		     'f:o' =>
"FC          = \@FC\@
FFLAGS      = \@FFLAGS\@
F77_COMPILE = \$(FC) \$(FFLAGS) \$(F77INCLUDES)",
	    'cxx:o' => "CXX         = \@CXX\@
CXXFLAGS    = \@CXXFLAGS\@
CXX_COMPILE = \$(CXX) \$(DEFS) \$(INCLUDES) \$(CXXFLAGS) \$(CPPFLAGS)",
	    'cpp:o' => "CXX         = \@CXX\@
CXXFLAGS    = \@CXXFLAGS\@
CXX_COMPILE = \$(CXX) \$(DEFS) \$(INCLUDES) \$(CXXFLAGS) \$(CPPFLAGS)",
	    'cs:o'  => "CSHARP      = \@CSHARP\@
CSFLAGS     = \@CSFLAGS\@
CSHARP_COMPILE = \$(CSHARP) \$(DEFS) \$(INCLUDES) \$(CSFLAGS) \$(CPPFLAGS)",
	    's:o'   => "AS          = \@AS\@
AS_COMPILE  = \$(AS)",
	    'f90:o'   => "F90          = \@F90\@
F90FLAGS    = \@F90FLAGS\@
F90_COMPILE = \$(F90) \$(F90FLAGS) \$(F90INCLUDES)",
# Similarly, but for shared-library versions
# (We may want to put these into a separate shlib.pl module)
	    'c:lo'   => "CC_SHL      = \@CC_SHL\@
C_COMPILE_SHL   = \$(CC_SHL) \$(DEFS) \$(INCLUDES) \$(CFLAGS) \$(CPPFLAGS)",
	    'f:lo'   => "FC_SHL      = \@FC_SHL\@
F77_COMPILE_SHL = \$(FC_SHL) \$(FFLAGS)",
	    'cxx:lo' => "CXX_SHL     = \@CXX_SHL\@
CXX_COMPILE_SHL = \$(CXX_SHL) \$(DEFS) \$(INCLUDES) \$(CXXFLAGS) \$(CPPFLAGS)",
	    'cpp:lo' => "CXX_SHL     = \@CXX_SHL\@
CXX_COMPILE_SHL = \$(CXX_SHL) \$(DEFS) \$(INCLUDES) \$(CXXFLAGS) \$(CPPFLAGS)",

		     'tex:dvi' => "",
		     );

# These are in one hash to simplify lookup.  Modules for new languages/features
# should update this hash (this is an Exported Variable)
%SOURCEBuildRule = (
# Source to object rules		    
		    'c:o' => "\$(C_COMPILE) -c \$<",
		    'f:o' => "\$(F77_COMPILE) -c \$<",
		    'cxx:o' => "\$(CXX_COMPILE) -c \$<",
		    's:o' => "\$(AS_COMPILE) \$<",
		    'f90:o' => "\$(F90_COMPILE) -c \$<",

# Program from source rules
		    'c:' => "\$(C_COMPILE) -o \$* \$< \$(LDFLAGS) \$(LIBS)",
		    'f:' => "\$(F77_COMPILE) -o \$* \$< \$(LDFLAGS) \$(LIBS)",
		    'f90:' => "\$(F90_COMPILE) -o \$* \$< \$(LDFLAGS) \$(LIBS)",
		    'cxx:' => "\$(CXX_COMPILE) -o \$* \$< \$(LDFLAGS) \$(LIBS)",
# LaTeX definitions
		    'tex:dvi' => "latex $<",
		    'dvi:pdf' => "dvipdfm $<",
		    'dvi:ps'  => "dvips $<",

# Assembler *output* definitions
		    'c:s' => "\$(C_COMPILE) -S \$<",
		    'f:s' => "\$(F77_COMPILE) -S \$<",

# C preprocessor *output* definitions.  We don't use cpp as the 
# output suffix to avoid conflicts with C++ 
		    'c:txt' => "\$(CPP) \$(INCLUDES) \$< >\$*.txt",
		    'F:txt' => "\$(CPP) \$(INCLUDES) \$<  >\$*.txt",
		    'cxx:txt' => "\$(CXXCPP) \$(INCLUDES) \$< >\$*.txt",

# Shared object definitions  (from simplemake; 
	      'c:lo' => "\$(C_COMPILE_SHL) -c \$<
	\@mv -f \$*.o \$*.lo", 
	      'f:lo' => "\$(F77_COMPILE_SHL) -c \$<
	\@mv -f \$*.o \$*.lo",
	      'cxx:lo' => "\$(CXX_COMPILE_SHL) -c \$<
	\@mv -f \$*.o \$*.lo",
	      'cpp:lo' => "\$(CXX_COMPILE_SHL) -c \$<
	\@mv -f \$*.o \$*.lo",
	      'f90:lo' => "\$(F90_COMPILE_SHL) -c \$<
	\@mv -f \$*.o \$*.lo",
# This has an MPICH-specific definition that we should eventually generalize
              'c:pf' => "\${C_COMPILE} -c -DMPICH_PROFILE_PMPI \$< 
	\@mv -f \$*.o \$*.pf",

		    );

#
# Given a string of blank-separated files, return the extension that 
# they ALL use (.o files may be ignored).  If multiple extensions are seen,
# return an empty string
sub SOURCEFindExtensions {
    my @filelist = split(/\s+/,$_[0]);
    my $ext = "";
    foreach my $file (@filelist) {
	if ($file =~ /.*\.(\w*)$/) {
	    my $extN = $1;
	    if ($extN eq "o") { next; }
	    if ($ext ne "") {
		if ($ext ne $extN) {
		    return "";
		}
	    }
	    else {
		$ext = $extN;
	    }
	}
    }
    return $ext;
}

# ----------------------------------------------------------------------------
# Routines exported to other modules
# Return the source files for a particular class of targets.
# Example:
#    SOURCEGetSourceFiles( "LIB" ) returns the files for all LIB_SOURCE
#    targets.
sub SOURCEGetSourceFiles {
    my $sourceClass = $_[0];

    my $hashname = "SOURCE_$sourceClass";
    my %filelist = ();
    # Make sure that we name each source file only once by placing it
    # into a hash.
    foreach $key (keys(%$hashname)) {
	foreach $file (split(/\s+/,$$hashname{$key})) {
	    $filelist{$file} = 1;
	}
    }
    # 
    return join(" ", sort(keys(%filelist)) );
    
}

# Return the source files for a particular extension
# Example
#    SOURCEGetSourceFilesByType( "c" ) returns all C files (use "h" to 
#    get header files)
sub SOURCEGetSourceFilesByType {
    my $ext = $_[0];
    
    # FIXME: need to implement.  Keep track on reading?
    return "";
}

# Given a blank-separate list of sources, convert to object file names
sub SOURCESToObjects {
    my $sources = $_[0];

    $objects = "";
    foreach my $source (split(/\s+/,$sources)) {
	if ($source =~ /^(.*)\.([^\.]*)$/) {
	    my $basename = $1;
	    my $ext      = $2;
	    # Check for known extensions
	    $objects .= "$basename.o ";
	}
    }
    return $objects;
}
# ---------------------------------------------------------------------------
# FIXME: Still to do is the support for profile libraries.
# ---------------------------------------------------------------------------
return 1;
