# ---------------------------------------------------------------------------
# -*- Mode:Perl; -*-
#DOC
# subdirs.pl - Provides support for specifying subdirectories
# 
# NOTSIMPLEMAKE_SUBDIRS - special case to make directories that are not
#   managed by simplemake.
#
# Functions Exported:
#   SUBDIRSGetManaged
#   SUBDIRSRecursiveOp
#ENDDOC
# Directory Control
#
# In addition
@SUBDIRS_ = ();
%SUBDIRSKnown = ();
#
# We need a way to mark a directory to be included in recursive operations
# but NOT in simplemake processing.  See SUBDIRSGetManaged; 
%UnmanagedDir = ();
#
sub ActionSUBDIRS {
    my ($sub,$term,$value,$filename,$linecount) = @_;

    if ($sub eq "") {
	@SUBDIRS_ = split(/\s+/,$value);
    }
    elsif ($sub eq "NOTSIMPLEMAKE") {
	for my $dir (split(/\s+/,$value)) {
	    UnmanagedDir{$dir} = 1;
	}
    }
    else {
	my $dirclass = "SUBDIRS_$sub";
	print "Creating dirname $dirclass\n" if $gDebug;
	@$dirclass = split(/\s+/,$value);
	$SUBDIRSKnown{$dirclass} = $value;
    }
    
    # If there are any subdirs, register MAKE for possible recursive 
    # output
    &ConfigureRegister( "MAKE", "\@MAKE\@" );
}
sub ClearSUBDIRS {
    @SUBDIRS_ = ();
    foreach my $key (keys(%SUBDIRSKnown)) {
	print "removing subdirs $key\n" if $gDebug;
	@$key = ();
	undef(@$key);
    }
    %SUBDIRSKnown = ();
    %UnmanagedDir = ();
}
#
# Return an array of the subdirectories that should be managed by simplemake
sub SUBDIRSGetManaged {
    my @alldirs;
    my @retdirs;
    my %uniquedir = ();
    @alldirs = @SUBDIRS_;
    foreach my $key (keys(%SUBDIRSKnown)) {
	@alldirs = (@alldirs, @$key);
    }
    # Remove any duplicates without changing the 
    # order.  Also remove any "unmanaged" directories
    @retdirs = ();
    foreach my $dir (@alldirs) {
	if (defined($UnmanagedDir{$dir})) {
	    next;
	}
	if (!defined($uniquedir{$dir})) {
	    $retdirs[$#retdirs+1] = $dir;
	    $uniquedir{$dir} = 1;
	}
    }
	
    return @retdirs;
}

# Register name
$KnownActions[$#KnownActions+1] = "SUBDIRS";

# ---------------------------------------------------------------------------
# Recursive operations
# SUBDIRSRecursiveOp( fd, target [, subdirectory class ] )
# Dependencies:
#    Relies on the directory values that may have been saved by the
#    SUBDIRS targets
#
# A "." in the subdirs path invokes target-local if defined
$gRecursiveThreshold = 3;
sub SUBDIRSRecursiveOp {
    my ($FD, $op, $dirclass) = @_;
    my $dirname = "SUBDIRS_";

    if ($dirclass) { $dirname .= $dirclass; }

    if (defined(@$dirname)) {
	print "Using dirname $dirname\n" if $gDebug;
	# Determine whether we should simply list the directories
	# or use a loop.  We must use a loop if:
	#   some items are either make or autoconf variables
	#   (in which case they may represent 0 or more than 1 directory)
	my $dircount = 0;
	my $hasDot = 0;
	foreach my $dir (@$dirname) {
	    if ($dir =~ /\$/ || $dir =~ /\@/) {
		$dircount += 1000;
	    }
	    if ($dir eq ".") {
		$hasDot = 1;
	    }
	    $dircount++;
	}

	# generate the output.  Watch for . if necessary
	if ($dircount < $gRecursiveThreshold) {
	    foreach my $dir (@$dirname) {
		if ($dir eq ".") {
		    if (defined($MakeTargets{"$op-local"})) {
			print $FD "\t\$(MAKE) $op-local\n";
		    }
		}
		else {
		    print $FD "\t(cd $dir && \$(MAKE) $op)\n";
		}
	    }
	}
	else {
	    if ($dircount >= 1000) {
		# handle possible empty lists by adding a - as
		# a sentinal.
		print $FD "\tfor dir in " . join( " ", @$dirname ) . " - ; do \\\n";
		if ($hasDot) {
		    print $FD "\t\tif [ \$\$dir = \".\" ] ; then \\\n";
		    if (defined($MakeTargets{"$op-local"})) {
			print $FD "\t\t    \$(MAKE) $op-local ; \\\n";
		    }
		    print $FD "\t\t    continue ; fi \\\n";
		}
		print $FD "\t\tif [ \$\$dir = \"-\" ] ; then break ; fi \\\n";
		print $FD "\t\t(cd \$\$dir && \$(MAKE) $op) ; done\n";
	    }
	    else {
		print $FD "\tfor dir in " . join( " ", @$dirname ) . " ; do \\\n";
		if ($hasDot) {
		    print $FD "\t\tif \$\$dir = \".\" ] ; then \\\n";
		    if (defined($MakeTargets{"$op-local"})) {
			print $FD "\t\t    \$(MAKE) $op-local ; \\\n";
		    }
		    print $FD "\t\t    continue ; fi\n";
		}
		print $FD "\t\t(cd \$\$dir && \$(MAKE) $op) ; done\n";
	    }
	}
    }
    
    # If there is no explicit dot target, add an implicit one to the end
    if (!$hasDot && defined($MakeTargets{"$op-local"})) {
	print $FD "\t\$(MAKE) $op-local\n";
    }
    
}

return 1;
